#ifndef __NF_SET_IPHASH_COMMON_H
#define __NF_SET_IPHASH_COMMON_H

#include <linux/netfilter/nf_set.h>

#if BITS_PER_LONG < 64
#define IPHASH_LEN	3
#else
#define IPHASH_LEN	4
#endif

struct iphash {
	struct iphash *next;
	nf_set_ip_t ip[IPHASH_LEN];
};

struct __nf_set_iphash {
	uint32_t hashsize;		/* hash size */
	uint32_t limit;			/* max number of IP addresses */
	uint32_t count;			/* current number of IP addresses */
	uint32_t initval;		/* initval for jhash_1word */
	struct iphash *members[0];	/* the iphash proper */
};

#ifdef __KERNEL__
static inline int
__iphash_testip(struct __nf_set_iphash *map, nf_set_ip_t ip)
{
	struct iphash *elem;
	__u32 id;
	int i;

	id = jhash_1word(ip, map->initval) % map->hashsize;
	DP("hash key: %u", id);
	elem = map->members[id];
	while (elem) {
		DP("try bucket %p", elem);
		for (i = 0; i < IPHASH_LEN; i++)
			if (elem->ip[i] == ip) {
				DP("found at %u", i);
				return 1;
			}
		elem = elem->next;
	}
	return 0;
}

static inline struct iphash *
__iphash_iphash_alloc(unsigned int cache_id)
{
	struct iphash *elem = (struct iphash *)
		nf_set_cache_alloc(cache_id, GFP_ATOMIC);
	
	if (elem)
		memset(elem, 0, sizeof(struct iphash));
	
	return elem;
}

static inline int
__iphash_addip(unsigned int cache_id, 
	       struct __nf_set_iphash *map, nf_set_ip_t ip)
{
	struct iphash *elem;
	__u32 id;
	int i;

	if (map->count >= map->limit)
		return -EOVERFLOW;
	
	id = jhash_1word(ip, map->initval) % map->hashsize;
	DP("hash key: %u", id);
	if (!map->members[id]) {
		map->members[id] = __iphash_iphash_alloc(cache_id);
		if (!map->members[id])
			return -ENOMEM;
		DP("create bucket %u (%p)", id, map->members[id]);
	}
	elem = map->members[id];
	while (1) {
		for (i = 0; i < IPHASH_LEN; i++) {
			if (elem->ip[i] == ip)
				return -EEXIST;
			if (!elem->ip[i]) {
				elem->ip[i] = ip;
				map->count++;
				return 0;
			}
		}
		if (!elem->next) {
			elem->next = __iphash_iphash_alloc(cache_id);
			if (!elem->next)
				return -ENOMEM;
			DP("create elem %p", elem->next);
		}
		elem = elem->next;
	}
	return 0;
}

static inline struct iphash *
find_last(struct iphash *elem, int *i)
{
	struct iphash *prev;
	
	if (!elem->next)
		return NULL;
		
	while (elem->next) {
		prev = elem;
		elem = elem->next;
	}
		
	for (*i = IPHASH_LEN - 1; *i >= 0; (*i)--) {
		if (elem->ip[*i])
			return prev;
	}
	BUG();
	return NULL;
}

static inline int
__iphash_delip(unsigned int cache_id,
	       struct __nf_set_iphash *map, nf_set_ip_t ip)
{
	struct iphash *elem, *prev, *pprev;
	__u32 id;
	int i, j;

	id = jhash_1word(ip, map->initval) % map->hashsize;
	DP("hash key: %u", id);
	prev = elem = map->members[id];
	while (elem) {
		for (i = 0; i < IPHASH_LEN; i++) {
			if (elem->ip[i] == ip) {
				DP("found at %u", i);
				map->count--;
				pprev = find_last(elem, &j);
				if (pprev) {
					DP("found prev %p", pprev);
					elem->ip[i] = pprev->next->ip[j];
					pprev->next->ip[j] = 0;
					prev = pprev;
					elem = prev->next;
				} else
					elem->ip[i] = 0;

				for (j = 0, i = 0; i < IPHASH_LEN; i++)
					if (!elem->ip[i])
						j++;
				if (j == IPHASH_LEN) {
					DP("freeing %p", elem);
					if (elem == map->members[id])
						map->members[id] = NULL;
					else
						prev->next = NULL;

					nf_set_cache_free(cache_id, elem);
				}
				return 0;
			}
		}
		prev = elem;
		elem = elem->next;
	}
	return -EEXIST;
}

static inline void 
__iphash_destroy(unsigned int cache_id, struct __nf_set_iphash *map)
{
	struct iphash *elem, *tmp;
	uint32_t i;

	for (i = 0; i < map->hashsize; i++) {
		elem = map->members[i];
		while (elem) {
			tmp = elem->next;
			nf_set_cache_free(cache_id, elem);
			elem = tmp;
		}
	}
}

static inline void 
__iphash_flush(unsigned int cache_id, struct __nf_set_iphash *map)
{
	struct iphash *elem, *tmp;
	uint32_t i;

	for (i = 0; i < map->hashsize; i++) {
		elem = map->members[i];
		while (elem) {
			tmp = elem->next;
			nf_set_cache_free(cache_id, elem);
			elem = tmp;
		}
		map->members[i] = NULL;
	}
	map->count = 0;
}

static inline void 
__iphash_list_members(const struct __nf_set_iphash *map, void *data)
{
	struct iphash *elem;
	nf_set_ip_t i, k;
	int j;

	k = 0;
	for (i = 0;  i < map->hashsize; i++) {
		elem = map->members[i];
		while (elem) {
			for (j = 0; j < IPHASH_LEN; j++)
				if (elem->ip[j])
					((nf_set_ip_t *)data)[k++] =
						elem->ip[j];
			elem = elem->next;
		}
	}
}
#endif /* __KERNEL__ */
#endif /* __NF_SET_IPHASH_COMMON_H */
