/* Copyright (C) 2000-2002 Joakim Axelsson <gozem@linux.nu>
 *                         Patrick Schaaf <bof@bof.de>
 * Copyright (C) 2003-2007 Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  
 */

/* Kernel module for IP set management */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/vmalloc.h>

#include <linux/netfilter/nfnetlink.h>
#include <linux/netfilter/nf_set_netlink.h>
#include <linux/netfilter/nf_set.h>

#define CONFIG_NF_SET_MAX	256		/* XXXX: Remove before release! */

static struct list_head set_type_list;		/* all registered set types */
static struct nf_set **nf_set_list; 		/* all individual sets */
static DEFINE_RWLOCK(nf_set_lock);		/* protects the lists */
static nf_set_id_t nf_set_max = CONFIG_NF_SET_MAX;
static unsigned int nf_set_hash_random;		/* random seed */

static int max_sets;
module_param(max_sets, int, 0600);
MODULE_PARM_DESC(max_sets, "maximal number of sets");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>");
MODULE_DESCRIPTION("module implementing core IP set support");
MODULE_ALIAS_NFNL_SUBSYS(NFNL_SUBSYS_NFSET);

/*
 * Sets are identified by the index in nf_set_list.
 * The index may change by swapping two sets.
 */

static inline void
__nf_set_get(struct nf_set *set)
{
	atomic_inc(&set->ref);
}

static inline void
__nf_set_put(struct nf_set *set)
{
	atomic_dec(&set->ref);
}

static inline void
__nf_set_get_id(nf_set_id_t index)
{
	atomic_inc(&nf_set_list[index]->ref);
}

static inline void
__nf_set_put_id(nf_set_id_t index)
{
	atomic_dec(&nf_set_list[index]->ref);
}

/*
 * Kernel API functions
 */

int
nf_set_testip_kernel(nf_set_id_t index,
		     const struct sk_buff *skb,
		     const u_int32_t flags)
{
	struct nf_set *set;
	int ret;
	
	set = nf_set_list[index];
	NF_SET_ASSERT(set);
	DSET("set %s, index %u", set->name, index);
	read_lock_bh(&set->lock);
	ret = set->type->testip_kernel(set, skb, flags);
	read_unlock_bh(&set->lock);

	return ret;
}

void
nf_set_addip_kernel(nf_set_id_t index,
		    const struct sk_buff *skb,
		    const u_int32_t flags)
{
	struct nf_set *set;
	int ret;

	set = nf_set_list[index];
	NF_SET_ASSERT(set);
	DSET("set %s, index %u", set->name, index);
retry:
	write_lock_bh(&set->lock);
	ret = set->type->addip_kernel(set, skb, flags);
	write_unlock_bh(&set->lock);
	if (ret == -EAGAIN
	    && set->type->retry
	    && (ret = set->type->retry(set)) == 0)
	    	goto retry;
}

void
nf_set_delip_kernel(nf_set_id_t index,
		    const struct sk_buff *skb,
		    const u_int32_t flags)
{
	struct nf_set *set;
	int ret;

	set = nf_set_list[index];
	NF_SET_ASSERT(set);
	DSET("set %s, index %u", set->name, index);
	write_lock_bh(&set->lock);
	ret = set->type->delip_kernel(set, skb, flags);
	write_unlock_bh(&set->lock);
}

/* Register and deregister settype */

#define MATCH_NAME(a,b)		!strncmp(a, b, NF_SET_MAXNAMELEN - 1)
#define NORMALIZE_NAME(name)	name[NF_SET_MAXNAMELEN-1] = '\0'

static inline struct nf_set_type *
find_set_type(const char *name, int family)
{
	struct nf_set_type *set_type;

	list_for_each_entry(set_type, &set_type_list, list)
		if (MATCH_NAME(set_type->typename, name)
		    && set_type->family == family)
			return set_type;
	return NULL;
}

int 
nf_set_register_set_type(struct nf_set_type *set_type)
{
	int ret = 0;
	
	if (set_type->protocol_version != NF_SET_PROTOCOL_VERSION) {
		nf_set_printk("'%s' uses wrong protocol version %u (want %u)",
			      set_type->typename,
			      set_type->protocol_version,
			      NF_SET_PROTOCOL_VERSION);
		return -EINVAL;
	}

	write_lock_bh(&nf_set_lock);
	if (find_set_type(set_type->typename, set_type->family)) {
		/* Duplicate! */
		nf_set_printk("%s with family %u already registered!", 
			      set_type->typename, set_type->family);
		ret = -EINVAL;
		goto unlock;
	}
	if (!try_module_get(THIS_MODULE)) {
		ret = -EFAULT;
		goto unlock;
	}
	list_add(&set_type->list, &set_type_list);
	DSET("%s (%u) registered.", set_type->typename, set_type->family);
   unlock:
	write_unlock_bh(&nf_set_lock);
	return ret;
}

void
nf_set_unregister_set_type(struct nf_set_type *set_type)
{
	write_lock_bh(&nf_set_lock);
	if (!find_set_type(set_type->typename, set_type->family)) {
		nf_set_printk("%s (%u) not registered?",
			      set_type->typename, set_type->family);
		goto unlock;
	}
	list_del(&set_type->list);
	module_put(THIS_MODULE);
	DSET("%s (%u) unregistered.", set_type->typename, set_type->family);
   unlock:
	write_unlock_bh(&nf_set_lock);
}

/*
 * Userspace routines
 */
 
/*
 * command		userspace		kernel
 * CREATE		CREATE			success
 * DESTROY/FLUSH	DESTROY/FLUSH		success
 * RENAME/SWAP		RENAME/SWAP		success
 * ADD/DEL/TEST		HEADER			create-data (create)
 *			ADD/DEL/TEST		success
 * LIST			LIST			set-names
 *			SET_HEADER		create-data (full)
 *			LIST_SET		set-elements
 */

/*
 * Find set by name, reference it once. The reference makes sure the
 * thing pointed to, does not go away under our feet. Drop the reference
 * later, using nf_set_put().
 */
nf_set_id_t
nf_set_get(const char *name)
{
	nf_set_id_t i, index = NF_SET_INVALID_ID;
	
	read_lock_bh(&nf_set_lock);
	for (i = 0; i < nf_set_max; i++) {
		if (nf_set_list[i] != NULL
		    && MATCH_NAME(nf_set_list[i]->name, name)) {
			__nf_set_get_id(i);
			index = i;
			break;
		}
	}
	read_unlock_bh(&nf_set_lock);
	return index;
}

/*
 * If the given set pointer points to a valid set, decrement
 * reference count by 1. The caller shall not assume the index
 * to be valid, after calling this function.
 */
void nf_set_put(nf_set_id_t index)
{
	read_lock_bh(&nf_set_lock);
	if (nf_set_list[index])
		__nf_set_put_id(index);
	read_unlock_bh(&nf_set_lock);
}

/*
 * Add, del and test IP in a set
 *
 * Return codes:
 *	EINVAL:	invalid core input parameters
 *	ENOENT:	no such set
 *	EEXIST: entry is already added/not in set
 *
 *	+ set type specific return codes
 */

static const int nfset_adt_min[NFSET_ADT_MAX] = {
	[NFSET_ADT_INDEX-1] = sizeof(nf_set_id_t),
	[NFSET_ADT_DATA-1] = -1,
};

#define ADT_CHECK(cda, set) 						\
({									\
	nf_set_id_t index;						\
	/* Check attributes and sizes */				\
	if (nfattr_invalid_size(cda, NFSET_ADT_MAX, nfset_adt_min))	\
		return -EINVAL;						\
									\
	index = *(nf_set_id_t *) NFA_DATA(cda[NFSET_ADT_INDEX-1]);	\
									\
	read_lock_bh(&nf_set_lock);					\
	set = nf_set_list[index];					\
	if (!set) {							\
		ret = -ENOENT;						\
		goto out;						\
	}								\
})

static int
nf_set_addip(struct sock *ctnl, struct sk_buff *skb,
	     struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	struct nf_set *set = NULL;
	int ret = 0;

	ADT_CHECK(cda, set);
	
	do {
		write_lock_bh(&set->lock);
		ret = set->type->addip(set, cda[NFSET_ADT_DATA-1]);
		write_unlock_bh(&set->lock);
	} while (ret == -EAGAIN
		 && set->type->retry
		 && (ret = set->type->retry(set)) == 0);

   out:
	read_unlock_bh(&nf_set_lock);
	return ret;
}

static int
nf_set_delip(struct sock *ctnl, struct sk_buff *skb,
	     struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	struct nf_set *set = NULL;
	int ret = 0;

	ADT_CHECK(cda, set);
	
	write_lock_bh(&set->lock);
	ret = set->type->delip(set, cda[NFSET_ADT_DATA-1]);
	write_unlock_bh(&set->lock);
	
   out:
	read_unlock_bh(&nf_set_lock);
	return ret;
}

static int
nf_set_testip(struct sock *ctnl, struct sk_buff *skb,
	      struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	struct nf_set *set = NULL;
	int ret = 0;

	ADT_CHECK(cda, set);
	
	read_lock_bh(&set->lock);
	ret = set->type->testip(set, cda[NFSET_ADT_DATA-1]);
	read_unlock_bh(&set->lock);
	
   out:
	read_unlock_bh(&nf_set_lock);
	return ret;
}

/*
 * Create a set
 *
 * Return codes:
 *	EINVAL:	invalid core input parameters
 *	ENOMEM:	not enough memory
 *	ENOENT:	no such set type
 *	EFAULT: could not get set type module
 *	EEXIST: set already exists
 *	EBADSLT: set table is full
 *
 *	+ set type specific return codes
 */
	
static struct nf_set_type *
find_set_type_rlock(const char *typename, int family)
{
	struct nf_set_type *type;
	
	read_lock_bh(&nf_set_lock);
	type = find_set_type(typename, family);
	if (type == NULL)
		read_unlock_bh(&nf_set_lock);

	return type;
}

static const int nfset_create_min[NFSET_CREATE_MAX] = {
	[NFSET_CREATE_NAME-1] = NF_SET_MAXNAMELEN,
	[NFSET_CREATE_TYPENAME-1] = NF_SET_MAXNAMELEN,
	[NFSET_CREATE_FAMILY-1] = sizeof(int),
	[NFSET_CREATE_DATA-1] = -1,
};

static int
nf_set_create(struct sock *ctnl, struct sk_buff *skb,
	      struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	char *name, *typename;
	struct nf_set *set;
	nf_set_id_t i, index = NF_SET_INVALID_ID;
	int family, ret = 0;
	
	/* Check attributes and sizes, except type specific data size */
	if (nfattr_invalid_size(cda, NFSET_CREATE_MAX, nfset_create_min))
		return -EINVAL;
	
	name = (char *) NFA_DATA(cda[NFSET_CREATE_NAME-1]);
	typename = (char *) NFA_DATA(cda[NFSET_CREATE_TYPENAME-1]);
	family = *(int *) NFA_DATA(cda[NFSET_CREATE_FAMILY-1]);
	NORMALIZE_NAME(name);
	NORMALIZE_NAME(typename);

	DSET("set: %s, type: %s", name, typename);
	/*
	 * First, and without any locks, allocate and initialize
	 * a normal base set structure.
	 */
	set = kmalloc(sizeof(struct nf_set), GFP_KERNEL);
	if (!set)
		return -ENOMEM;
	set->lock = RW_LOCK_UNLOCKED;
	strncpy(set->name, name, NF_SET_MAXNAMELEN);
	atomic_set(&set->ref, 0);

	/*
	 * Next, take the &nf_set_lock, check that we know the type,
	 * and take a reference on the type, to make sure it
	 * stays available while constructing our new set.
	 *
	 * After referencing the type, we drop the &nf_set_lock,
	 * and let the new set construction run without locks.
	 */
	set->type = find_set_type_rlock(typename, family);
	if (set->type == NULL) {
		/* Try loading the module */
		char modulename[NF_SET_MAXNAMELEN + strlen("nf_set_") + 1];
		strcpy(modulename, "nf_set_");
		strcat(modulename, typename);
		DSET("try to load %s", modulename);
		request_module(modulename);
		set->type = find_set_type_rlock(typename, family);
	}
	if (set->type == NULL) {
		nf_set_printk("no set type '%s', set '%s' not created",
			      typename, name);
		ret = -ENOENT;
		goto out;
	}
	if (!try_module_get(set->type->me)) {
		read_unlock_bh(&nf_set_lock);
		ret = -EFAULT;
		goto out;
	}
	read_unlock_bh(&nf_set_lock);

	/*
	 * Without holding any locks, create private part.
	 */
	ret = set->type->create(set, cda[NFSET_CREATE_DATA-1]);
	if (ret != 0)
		goto put_out;

	/* BTW, ret==0 here. */

	/*
	 * Finally! Add our shiny new set to the list, and be done.
	 */
	write_lock_bh(&nf_set_lock);
	for (i = 0; i < nf_set_max; i++) {
		if (nf_set_list[i] == NULL) {
			if (index == NF_SET_INVALID_ID)
				index = i;
		} else if (MATCH_NAME(nf_set_list[i]->name, name)) {
			ret = -EEXIST;
			goto cleanup;
		}
	}
	if (index != NF_SET_INVALID_ID) {
		DSET("create: '%s' created with index %u", 
			set->name, index);
		nf_set_list[index] = set;
		write_unlock_bh(&nf_set_lock);
		return ret;
	}
	nf_set_printk("Cannot create set '%s', max number of sets %u already created",
		set->name, nf_set_max);
	ret = -EBADSLT;
    cleanup:
	write_unlock_bh(&nf_set_lock);
	set->type->destroy(set);
    put_out:
	module_put(set->type->me);
    out:
	kfree(set);
	return ret;
}

/*
 * Destroy a set - or all sets
 * Sets must not be referenced/used.
 *
 * Return codes:
 *	EINVAL:	invalid core input parameters
 *	EBUSY:	set is in use
 *	ENOENT: set does not exist
 */
 
static const int nfset_list_min[NFSET_LIST_MAX] = {
	[NFSET_LIST_NAME-1] = NF_SET_MAXNAMELEN,
};

static inline int
__nf_set_destroy(struct nf_set *set)
{
	int ret;
	
	DSET("set: %s",  set->name);
	/* Must call it without holding any lock */
	ret = set->type->destroy(set);
	module_put(set->type->me);
	kfree(set);
	
	return ret;
}

static inline int
__nf_set_busy(struct nf_set *set)
{
	return (atomic_read(&set->ref) 
		|| (set->type->referenced && set->type->referenced(set)));
}

static int
nf_set_destroy(struct sock *ctnl, struct sk_buff *skb,
	       struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	struct nf_set *set = NULL;
	char *name;
	nf_set_id_t i;
	int ret = 0, res;
	
	if (nlh->nlmsg_flags & NLM_F_DUMP) {
		/* Destroy all not-in-use sets: no rollback! */
		nf_set_id_t j = 0;
	next_set:
		write_lock_bh(&nf_set_lock);
		for (i = j; i < nf_set_max; i++) {
			set = nf_set_list[i];
			if (set == NULL)
				continue;
			if (__nf_set_busy(set)) {
				ret = -EBUSY;
				continue;
			}
			nf_set_list[i] = NULL;
			j = i + 1;
			break;
		}
		write_unlock_bh(&nf_set_lock);
		if (set) {
			res = __nf_set_destroy(set);
			if (res == -EAGAIN) {
				/* Union of sets destroyed: restart */
				ret = j = 0;
				goto next_set;
			}
		}
		if (i < nf_set_max)
			goto next_set;
		return ret;
	}

	/* Check attributes and sizes */
	if (nfattr_invalid_size(cda, NFSET_LIST_MAX, nfset_list_min))
		return -EINVAL;
	
	name = (char *) NFA_DATA(cda[NFSET_LIST_NAME-1]);
	NORMALIZE_NAME(name);
	
	write_lock_bh(&nf_set_lock);
	for (i = 0; i < nf_set_max; i++) {
		set = nf_set_list[i];
		if (set != NULL && MATCH_NAME(set->name, name)) {
		    	if (__nf_set_busy(set))
		    		ret = -EBUSY;
		    	else
		    		nf_set_list[i] = NULL;
			break;
		}
	}
	write_unlock_bh(&nf_set_lock);

	if (set) 
		do {
			res = __nf_set_destroy(set);
		} while (res == -EAGAIN);
	else
		ret = -ENOENT;

	return ret;
}

/* 
 * Flush data in a set - or in all sets
 *
 * Return codes:
 *	EINVAL:	invalid core input parameters
 *	ENOENT: set does not exist
 */

static inline nf_set_id_t
__nf_set_find_rlock(const char *name)
{
	nf_set_id_t i, index = NF_SET_INVALID_ID;
	
	read_lock_bh(&nf_set_lock);
	for (i = 0; i < nf_set_max; i++)
		if (nf_set_list[i] 
		    && MATCH_NAME(nf_set_list[i]->name, name)) {
		    	index = i;
		    	break;
		}
	return index;
}

static inline void
__nf_set_flush(struct nf_set *set)
{
	DSET("set: %s",  set->name);
	write_lock_bh(&set->lock);
	set->type->flush(set);
	write_unlock_bh(&set->lock);
}

static int
nf_set_flush(struct sock *ctnl, struct sk_buff *skb,
	     struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	char *name;
	nf_set_id_t i, index;
	struct nf_set *set;
	int ret = 0;
	
	if (nlh->nlmsg_flags & NLM_F_DUMP) {
		/* Flush all sets */
		read_lock_bh(&nf_set_lock);
		for (i = 0; i < nf_set_max; i++) {
			set = nf_set_list[i];
			if (set == NULL)
				continue;
			__nf_set_flush(set);
		}
		read_unlock_bh(&nf_set_lock);
		return 0;
	}
	/* Check attributes and sizes */
	if (nfattr_invalid_size(cda, NFSET_LIST_MAX, nfset_list_min))
		return -EINVAL;
	
	name = (char *) NFA_DATA(cda[NFSET_LIST_NAME-1]);
	NORMALIZE_NAME(name);
	
	index = __nf_set_find_rlock(name);
	if (index != NF_SET_INVALID_ID)
		__nf_set_flush(nf_set_list[index]);
	else
		ret = -ENOENT;
	read_unlock_bh(&nf_set_lock);

	return ret;
}

/* 
 * Rename a set
 *
 * Return codes:
 *	EINVAL:	invalid core input parameters
 *	EBUSY:  set is in use
 *	EEXIST: set with new name exists
 *	ENOENT: set with old name does not exist
 */

static const int nfset_rename_min[NFSET_RENAME_MAX] = {
	[NFSET_RENAME_FROM-1] = NF_SET_MAXNAMELEN,
	[NFSET_RENAME_TO-1] = NF_SET_MAXNAMELEN,
};

static int
nf_set_rename(struct sock *ctnl, struct sk_buff *skb,
	      struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	char *from_name, *to_name;
	struct nf_set *set = NULL;
	nf_set_id_t i;
	int ret = 0;
	
	/* Check attributes and sizes */
	if (nfattr_invalid_size(cda, NFSET_RENAME_MAX, nfset_rename_min))
		return -EINVAL;
	
	from_name = (char *) NFA_DATA(cda[NFSET_RENAME_FROM-1]);
	to_name = (char *) NFA_DATA(cda[NFSET_RENAME_TO-1]);
	NORMALIZE_NAME(from_name);
	NORMALIZE_NAME(to_name);
	
	DSET("set: %s to %s",  from_name, to_name);
	write_lock_bh(&nf_set_lock);
	for (i = 0; i < nf_set_max; i++) {
		if (nf_set_list[i] == NULL)
			continue;
		if (set == NULL 
		    && MATCH_NAME(nf_set_list[i]->name, from_name)) {
			set = nf_set_list[i];
			if (atomic_read(&set->ref)) {
				ret = -EBUSY;
				goto unlock;
			}
		}
		if (MATCH_NAME(nf_set_list[i]->name, to_name)) {
			ret = -EEXIST;
			goto unlock;
		}
	}
	if (!set) {
		ret = -ENOENT;
		goto unlock;
	}
	strncpy(set->name, to_name, NF_SET_MAXNAMELEN);
    unlock:
	write_unlock_bh(&nf_set_lock);
	return ret;
}

/*
 * Swap two sets so that their content behind the names are swapped.
 *
 * Return codes:
 *	EINVAL:	invalid core input parameters
 *	ENOENT: either of the sets does not exist
 *	ENOEXEC: sets are different types
 */

static int
nf_set_swap(struct sock *ctnl, struct sk_buff *skb,
	    struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	char *from_name, *to_name;
	struct nf_set *from_set = NULL, *to_set = NULL;
	u_int32_t from_ref;
	nf_set_id_t i, from_index, to_index;
	int ret = 0;
	
	/* Check attributes and sizes */
	if (nfattr_invalid_size(cda, NFSET_RENAME_MAX, nfset_rename_min))
		return -EINVAL;
	
	from_name = (char *) NFA_DATA(cda[NFSET_RENAME_FROM-1]);
	to_name = (char *) NFA_DATA(cda[NFSET_RENAME_TO-1]);
	NORMALIZE_NAME(from_name);
	NORMALIZE_NAME(to_name);
	from_index = to_index = NF_SET_INVALID_ID;
	
	DSET("set: %s to %s",  from_name, to_name);

	write_lock_bh(&nf_set_lock);
	for (i = 0; 
	     i < nf_set_max && (from_set == NULL || to_set == NULL); 
	     i++) {
		if (nf_set_list[i] == NULL)
			continue;
		if (from_set == NULL 
		    && MATCH_NAME(nf_set_list[i]->name, from_name)) {
			from_set = nf_set_list[i];
			from_index = i;
		}
		if (to_set == NULL 
		    && MATCH_NAME(nf_set_list[i]->name, to_name)) {
			to_set = nf_set_list[i];
			to_index = i;
		}
	}
	if (!(from_set && to_set)) {
		ret = -ENOENT;
		goto unlock;
	}
	/* Family and features must not change.
	 * Latter is an artifical restriction. */
	if ((from_set->type->family != to_set->type->family)
	    || (from_set->type->features != to_set->type->features)) {
		ret = -ENOEXEC;
		goto unlock;
	}

	/* No magic here: ref munging protected by the lock */
	from_ref = atomic_read(&from_set->ref);

	strncpy(from_set->name, to_name, NF_SET_MAXNAMELEN);
	atomic_set(&from_set->ref, atomic_read(&to_set->ref));
	strncpy(to_set->name, from_name, NF_SET_MAXNAMELEN);
	atomic_set(&to_set->ref, from_ref);
	
	nf_set_list[from_index] = to_set;
	nf_set_list[to_index] = from_set;

unlock:
	write_unlock_bh(&nf_set_lock);
	return ret;
}

/*
 * List set header
 */

#define __nf_set_msg(skb, pid, seq, event, nowait)			\
({									\
	struct nlmsghdr *nlh;						\
	struct nfgenmsg *nfmsg;						\
									\
	event |= NFNL_SUBSYS_NFSET << 8;				\
	nlh = NLMSG_PUT(skb, pid, seq, event, sizeof(struct nfgenmsg));	\
	nfmsg  = NLMSG_DATA(nlh);					\
									\
	nlh->nlmsg_flags = (nowait && pid) ? NLM_F_MULTI : 0;		\
	nfmsg->nfgen_family = AF_INET;	/* FIXME */			\
	nfmsg->version = NFNETLINK_V0;					\
	nfmsg->res_id = 0;						\
})
	
static inline int
__nf_set_header(nf_set_id_t index, int full, 
		struct sk_buff *skb, u32 pid, u32 seq,
		int event, int nowait)
{
	struct nf_set *set = nf_set_list[index];
	int ret = 0;
	
	__nf_set_msg(skb, pid, seq, event, nowait);
	
	NFA_PUT(skb, NFSET_CREATE_NAME, NF_SET_MAXNAMELEN, set->name);
	NFA_PUT_SIMPLE(skb, NFSET_CREATE_INDEX, nf_set_id_t, index);	
	NFA_PUT(skb, NFSET_CREATE_TYPENAME, NF_SET_MAXNAMELEN, set->type->typename);
	NFA_PUT_SIMPLE(skb, NFSET_CREATE_FAMILY, int, set->type->family);
	
	if (full)		
		ret = set->type->list_header(skb, set);
		
	return (ret == 0 ? skb->len : ret);

nlmsg_failure:	
nfattr_failure:
	return -EFAULT;
}

static int
nf_set_header(struct sock *ctnl, struct sk_buff *skb,
	      struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	char *name;
	struct sk_buff *skb2;
	nf_set_id_t index;
	int ret = -ENOENT;
	
	/* Check attributes and sizes */
	if (nfattr_invalid_size(cda, NFSET_LIST_MAX, nfset_list_min))
		return -EINVAL;
	
	name = (char *) NFA_DATA(cda[NFSET_LIST_NAME-1]);
	NORMALIZE_NAME(name);
	
	skb2 = alloc_skb(NLMSG_GOODSIZE, GFP_KERNEL);
	if (!skb2)
		return -ENOMEM;
	
	index = __nf_set_find_rlock(name);
	if (index != NF_SET_INVALID_ID)
		ret = __nf_set_header(index, nlh->nlmsg_flags & NLM_F_DUMP,
				      skb,
				      NETLINK_CB(skb).pid,
				      nlh->nlmsg_seq,
				      NFSET_MSG_CREATE, 1);
	read_unlock_bh(&nf_set_lock);
	if (index == NF_SET_INVALID_ID || ret < 0)
		goto free;
	
	ret = netlink_unicast(ctnl, skb2, NETLINK_CB(skb).pid, MSG_DONTWAIT);
	if (ret < 0)
		goto out;
	
	return 0;

    free:
	kfree_skb(skb2);
    out:
	return ret;
}

/*
 * List set data
 */

/*
 * cb->args[5]:
 *
 *	args[0]: index
 *	args[1]: set pointer, if null, list header
 *	args[2-5]: reserved for set type specific parameters
 */

static inline int
__nf_set_list(nf_set_id_t index,
	      struct sk_buff *skb, struct netlink_callback *cb)
{
	struct nf_set *set;
	int ret = 0;
	
	/* First, list complete header data */
	read_lock_bh(&nf_set_lock);
	if (!cb->args[1]) {
		ret = __nf_set_header(index, 1, skb, NETLINK_CB(skb).pid,
				      cb->nlh->nlmsg_seq, NFSET_MSG_CREATE, 1);
		if (ret > 0)
			cb->args[1] = (unsigned long) nf_set_list[index];
		goto out;
	}

	/* Header is listed, continue with the set data */
	set = nf_set_list[index];
	/* Make sure set was not swapped under our feet */
	if (set != (struct nf_set *) cb->args[1]) {
		ret = -ENOEXEC;
		goto out;
	}
	read_lock_bh(&set->lock);
	ret = set->type->list_members(skb, cb, set);	
	read_unlock_bh(&set->lock);
	
    out:
	read_unlock_bh(&nf_set_lock);	
	return ret;
}

static int
nf_set_list_set_all(struct sk_buff *skb, struct netlink_callback *cb)
{
	nf_set_id_t index = NF_SET_INVALID_ID;
	int ret = 0;

	read_lock_bh(&nf_set_lock);
	/* For loop for empty set buckets */
	for (; cb->args[0] < nf_set_max; cb->args[0]++) {
		index = (nf_set_id_t) cb->args[0];
		if (!nf_set_list[index])
			continue;
		if (!cb->args[1])
			/* List new/next set */
			__nf_set_get_id(index);
		break;
	}
	read_unlock_bh(&nf_set_lock);
	if (index >= nf_set_max)
		return ret;

	cb->args[0] = index;
	ret = __nf_set_list(index, skb, cb);
	/* If current set is finished, release reference 
	 * and proceed with the next one */
	if (!cb->args[1])
		__nf_set_put_id((nf_set_id_t) cb->args[0]++);

	return ret;
}

static int
nf_set_list_named_set(struct sk_buff *skb, struct netlink_callback *cb)
{
	int ret = __nf_set_list(cb->args[0], skb, cb);
	/* If finished, release reference */
	if (!cb->args[1])
		__nf_set_put_id((nf_set_id_t) cb->args[0]);

	return ret;
}

static int
nf_set_list_set(struct sock *ctnl, struct sk_buff *skb,
		struct nlmsghdr *nlh, struct nfattr *cda[], int *errp)
{
	u32 rlen;
	
	if (nlh->nlmsg_flags & NLM_F_DUMP) {
		/* List all sets */
		*errp = netlink_dump_start(ctnl, skb, nlh,
					   nf_set_list_set_all,
					   NULL);
	} else {
		char *name;
		nf_set_id_t index;
		
		/* Check attributes and sizes */
		if (nfattr_invalid_size(cda, NFSET_LIST_MAX, nfset_list_min))
			return -EINVAL;
	
		name = (char *) NFA_DATA(cda[NFSET_LIST_NAME-1]);
		NORMALIZE_NAME(name);
	
		index = __nf_set_find_rlock(name);

		if (index == NF_SET_INVALID_ID) {
			read_unlock_bh(&nf_set_lock);
			return -ENOENT;
		}
		__nf_set_get_id(index);
		read_unlock_bh(&nf_set_lock);
		
		*errp = netlink_dump_init(ctnl, skb, nlh,
					  nf_set_list_named_set,
					  NULL,
					  1, index);
	}
	if (*errp)
		return -EINVAL;
	
	rlen = NLMSG_ALIGN(nlh->nlmsg_len);
	if (rlen > skb->len)
		rlen = skb->len;
	skb_pull(skb, rlen);
	return 0;
}

static struct nfnl_callback nfset_cb[NFSET_MSG_MAX] = {
	[NFSET_MSG_CREATE]	= { .call = nf_set_create,
				    .attr_count = NFSET_CREATE_MAX, },
	[NFSET_MSG_DESTROY]	= { .call = nf_set_destroy,
				    .attr_count = NFSET_LIST_MAX, },
	[NFSET_MSG_FLUSH]	= { .call = nf_set_flush,
				    .attr_count = NFSET_LIST_MAX, },
	[NFSET_MSG_RENAME]	= { .call = nf_set_rename,
				    .attr_count = NFSET_RENAME_MAX, },
	[NFSET_MSG_SWAP]	= { .call = nf_set_swap,
				    .attr_count = NFSET_RENAME_MAX, },
	[NFSET_MSG_HEADER]	= { .call = nf_set_header,
				    .attr_count = NFSET_LIST_MAX, },
	[NFSET_MSG_LIST]	= { .call = nf_set_list_set,
				    .attr_count = NFSET_LIST_MAX, },
	[NFSET_MSG_ADD]		= { .call = nf_set_addip,
				    .attr_count = NFSET_ADT_MAX, },
	[NFSET_MSG_DEL]		= { .call = nf_set_delip,
				    .attr_count = NFSET_ADT_MAX, },
	[NFSET_MSG_TEST]	= { .call = nf_set_testip,
				    .attr_count = NFSET_ADT_MAX, },
};

static struct nfnetlink_subsystem nfset_subsys = {
	.name		= "nfset",
	.subsys_id	= NFNL_SUBSYS_NFSET,
	.cb_count	= NFSET_MSG_MAX,
	.cb		= nfset_cb,
};

static int __init nf_set_init(void)
{
	int ret;

	get_random_bytes(&nf_set_hash_random, 4);
	if (max_sets)
		nf_set_max = max_sets;
	nf_set_list = vmalloc(sizeof(struct nf_set *) * nf_set_max);
	if (!nf_set_list) {
		nf_set_printk("Unable to create nf_set_list\n");
		return -ENOMEM;
	}
	memset(nf_set_list, 0, sizeof(struct nf_set *) * nf_set_max);

	INIT_LIST_HEAD(&set_type_list);

	ret = nfnetlink_subsys_register(&nfset_subsys);
	if (ret != 0) {
		nf_set_printk("Cannot register with netlink: %d", ret);
		vfree(nf_set_list);
		return ret;
	}
	return 0;
}

static void __exit nf_set_fini(void)
{
	/* There can't be any existing set here */
	nfnetlink_subsys_unregister(&nfset_subsys);
	vfree(nf_set_list);
	DSET("these are the famous last words");
}

EXPORT_SYMBOL_GPL(nf_set_register_set_type);
EXPORT_SYMBOL_GPL(nf_set_unregister_set_type);

EXPORT_SYMBOL_GPL(nf_set_get);
EXPORT_SYMBOL_GPL(nf_set_put);

EXPORT_SYMBOL_GPL(nf_set_addip_kernel);
EXPORT_SYMBOL_GPL(nf_set_delip_kernel);
EXPORT_SYMBOL_GPL(nf_set_testip_kernel);

module_init(nf_set_init);
module_exit(nf_set_fini);
