/* Copyright (C) 2003-2004 Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  
 */

/* Kernel module implementing an ip hash set */

#include <linux/module.h>
#include <linux/ip.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <linux/random.h>
#include <linux/vmalloc.h>
#include <net/ip.h>

#include <linux/netfilter/nf_set_malloc.h>
#include <linux/netfilter/nf_set_jhash.h>
#include <linux/netfilter/nf_set_iphash_common.h>
#include <linux/netfilter/nf_set_iphash.h>

#include <linux/netfilter/nfnetlink.h>
#include <linux/netfilter/nfnetlink_set.h>

static unsigned int cache_id;

static inline int
adt_data(struct nfattr *attr, int *errp)
{
	nf_set_ip_t ip;
	struct nfattr *tb[NFSET_ADT_DATA_MAX];

	nfattr_parse_nested(tb, NFSET_ADT_DATA_MAX, attr);

	if (!tb[NFSET_ADT_DATA_IPV4]) {
		*errp = -EINVAL;
		return 0;
	}
	ip = *(nf_set_ip_t *)NFA_DATA(tb[NFSET_ADT_DATA_IPV4-1]);

	return ip;
}

static inline int
__testip(struct nf_set *set, nf_set_ip_t ip)
{
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;
	nf_set_ip_t haship = ip & map->netmask;

	DP("set: %s, ip:%u.%u.%u.%u, mask:%u.%u.%u.%u %u.%u.%u.%u",
	   set->name, HIPQUAD(ip), HIPQUAD(map->netmask), HIPQUAD(haship));
	return __iphash_testip(&map->map, haship);
}

static int
testip(struct nf_set *set, struct nfattr *attr)
{
	int err = 0;

	nf_set_ip_t ip = adt_data(attr, &err);
	if (err < 0)
		return err;

	return __testip(set, ip);
}

static int
testip_kernel(struct nf_set *set, 
	      const struct sk_buff *skb,
	      const u_int32_t *flags)
{
	return __testip(set,
			ntohl(flags[0] & NFSET_SRC 
				? skb->nh.iph->saddr 
				: skb->nh.iph->daddr));
}

static inline int
__addip(struct nf_set *set, nf_set_ip_t ip)
{
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;
	nf_set_ip_t haship = ip & map->netmask;

	DP("set: %s, ip:%u.%u.%u.%u, mask:%u.%u.%u.%u %u.%u.%u.%u",
	   set->name, HIPQUAD(ip), HIPQUAD(map->netmask), HIPQUAD(haship));
	return __iphash_addip(cache_id, &map->map, haship);
}

static int
addip(struct nf_set *set, struct nfattr *attr)
{
	int err = 0;

	nf_set_ip_t ip = adt_data(attr, &err);
	if (err < 0)
		return err;

	return __addip(set, ip);
}

static int
addip_kernel(struct nf_set *set, 
	     const struct sk_buff *skb,
	     const u_int32_t *flags)
{
	return __addip(set,
		       ntohl(flags[0] & NFSET_SRC 
		       		? skb->nh.iph->saddr 
				: skb->nh.iph->daddr));
}

static inline int
__delip(struct nf_set *set, nf_set_ip_t ip)
{
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;
	nf_set_ip_t haship = ip & map->netmask;

	DP("set: %s, ip:%u.%u.%u.%u, mask:%u.%u.%u.%u %u.%u.%u.%u",
	   set->name, HIPQUAD(ip), HIPQUAD(map->netmask), HIPQUAD(haship));
	return __iphash_delip(cache_id, &map->map, haship);
}

static int
delip(struct nf_set *set, struct nfattr *attr)
{
	int err = 0;

	nf_set_ip_t ip = adt_data(attr, &err);
	if (err < 0)
		return err;

	return __delip(set, ip);
}

static int
delip_kernel(struct nf_set *set, 
	     const struct sk_buff *skb,
	     const u_int32_t *flags)
{
	return __delip(set,
		       ntohl(flags[0] & NFSET_SRC 
				? skb->nh.iph->saddr 
				: skb->nh.iph->daddr));
}

static int create(struct nf_set *set, struct nfattr *attr)
{
	struct nf_set_iphash *map;
	struct nfattr *tb[NFSET_IPHASH_MAX];
	int size, limit;

	nfattr_parse_nested(tb, NFSET_IPHASH_MAX, attr);

	if (!tb[NFSET_IPHASH_SIZE-1])
		return -EINVAL;
	size = *(__u32 *) NFA_DATA(tb[NFSET_IPHASH_SIZE-1]);

	if (size < 1) {
		nf_set_printk("hashsize too small");
		return -ENOEXEC;
	}

	if (!tb[NFSET_IPHASH_LIMIT-1])
		return -EINVAL;
	limit = *(__u32 *) NFA_DATA(tb[NFSET_IPHASH_LIMIT-1]);

	if (limit < 1) {
		nf_set_printk("limit too small");
		return -ENOEXEC;
	}

	map = kmalloc(sizeof(struct nf_set_iphash)
		      + size * sizeof(struct iphash *), GFP_KERNEL);
	if (!map) {
		DP("out of memory for %d bytes", 
		    sizeof(struct nf_set_iphash)
		    + size * sizeof(struct iphash *));
		return -ENOMEM;
	}
	memset(map, 0,
	       sizeof(struct nf_set_iphash)
	       + size * sizeof(struct iphash *));

	if (!tb[NFSET_IPHASH_NETMASK-1])
		return -EINVAL;

	map->netmask = *(__u32 *) NFA_DATA(tb[NFSET_IPHASH_NETMASK-1]);
	map->map.hashsize = size;
	map->map.limit = limit;
	get_random_bytes(&map->map.initval, 4);

	set->data = map;
	return 0;
}

static void destroy(struct nf_set *set)
{
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;

	__iphash_destroy(cache_id, &map->map);
	kfree(map);
	set->data = NULL;
}

static void flush(struct nf_set *set)
{
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;

	__iphash_flush(cache_id, &map->map);
}

static int list_header(struct sk_buff *skb, const struct nf_set *set)
{
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;

	NFA_PUT(skb, NFSET_IPHASH_SIZE, sizeof(map->map.hashsize),
		&map->map.hashsize);
	NFA_PUT(skb, NFSET_IPHASH_LIMIT, sizeof(map->map.limit),
		&map->map.limit);
	NFA_PUT(skb, NFSET_IPHASH_NETMASK, sizeof(map->netmask),
		&map->netmask);

	return 0;

nfattr_failure:
	return -1;
}

/* Number of members that fit in a netlink message */
#define nmembers ((NLMSG_GOODSIZE - NLMSG_SPACE(sizeof(struct nfgenmsg))) / 4)

static int list_members_init(struct nf_set *set, struct nf_set_dump *d)
{
	nf_set_ip_t *tmp;
	struct nf_set_iphash *map = (struct nf_set_iphash *) set->data;

	read_lock_bh(&set->lock);

	tmp = kmalloc(map->map.count * sizeof(nf_set_ip_t), GFP_ATOMIC);
	if (!tmp)
		return -ENOMEM;

	__iphash_list_members(&map->map, (void *)tmp);

	d->count = map->map.count;
	d->members = tmp;

	read_unlock_bh(&set->lock);

	return 0;
}

static int list_members(struct sk_buff *skb, struct nf_set_dump *d)
{
	int size;
	unsigned char *b;
	struct nlmsghdr *nlh;
	nf_set_ip_t *tmp = (nf_set_ip_t *) d->members;

	/* Nothing else to dump */
	if (!d->count)
		return 0;

	/* How many members are we going to put in a message? */
	if (d->count > nmembers)
		size = nmembers;
	else
		size = d->count;

	d->count = d->count - size;

	b = skb->tail;
	nlh = __nfset_nfnetlink_header(skb);
	NFA_PUT(skb, NFSET_ADT_DATA_IPV4, size * sizeof(nf_set_ip_t),
		&tmp[d->count]);
	nlh->nlmsg_len = skb->tail - b;

	return skb->len;

nfattr_failure:
	/* Shouldn't happen, hit bug */
	return 0;
}

static void list_members_done(struct nf_set_dump *d)
{
	kfree(d->members);
}

static struct nf_set_type nf_set_iphash = {
	.typename		= SETTYPE_NAME,
	.features		= NFSET_TYPE_IP | NFSET_TYPE_SINGLE,
	.subtype_base		= 0,
	.subtype_next		= 1,
	.me			= THIS_MODULE,
	.subtypes		= {
		[0] = {
			.create		= &create,
			.flags		= 0,
			.destroy	= &destroy,
			.flush		= &flush,
			.add		= &addip,
			.add_kernel	= &addip_kernel,
			.del		= &delip,
			.del_kernel	= &delip_kernel,
			.test		= &testip,
			.test_kernel	= &testip_kernel,
			.list_header	= &list_header,
			.list_members_init = &list_members_init,
			.list_members	= &list_members,
			.list_members_done = &list_members_done,
		},
	},
};

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>");
MODULE_DESCRIPTION("iphash type of IP sets");

static int __init init(void)
{
	int ret;
	
	ret = nf_set_cache_create(sizeof(struct iphash), &cache_id);
	if (ret) {
		printk(KERN_ERR "Unable to create slab cache for nf_set_iphash\n");
		return ret;
	}
	ret = nf_set_register_set_type(&nf_set_iphash);
	if (ret)
		nf_set_cache_destroy(cache_id);

	return ret;
}

static void __exit fini(void)
{	
	/* FIXME: possible race with nf_set_create() */
	nf_set_unregister_set_type(&nf_set_iphash);
	nf_set_cache_destroy(cache_id);
}

module_init(init);
module_exit(fini);
